//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

/*
 * CompressLZMH.c
 *
 *	RAM Consumption: INTERN_BUFFER_LENGTH + HUFF_LIST_LENGTH * 3 + 48 = 384 Byte
 *  This Code is under the GNU Lesser General Public License.
 *
 *  Created on: 28.09.2011
 *      Author: Martin Ringwelski
 */

#include "CompressLZMH.h"

#define CODE_BUFFER	32
typedef uint32_t	code_t;

#define HUFF_SYMBOL_NOT_FOUND	0xFFFF
//#define LZMH_FIRSTSEARCHHISTORY

typedef struct {
	uint8_t		symbol;
	uint16_t	count;
} HUFFLIST;

typedef struct {
	uint8_t	code;
	uint8_t	length;
} CODE_ELEMENT;

#define TREE_LENGHT	19
static const CODE_ELEMENT tree[TREE_LENGHT] = {
		{0x0F /*0b1111*/, 4},
		{0x0E /*0b1110*/, 4},
		{0x0D /*0b1101*/, 4},
		{0x0C /*0b1100*/, 4},
		{0x17 /*0b10111*/, 5},
		{0x16 /*0b10110*/, 5},
		{0x15 /*0b10101*/, 5},
		{0x14 /*0b10100*/, 5},
		{0x13 /*0b10011*/, 5},
		{0x25 /*0b100101*/, 6},
		{0x24 /*0b100100*/, 6},
		{0x23 /*0b100011*/, 6},
		{0x22 /*0b100010*/, 6},
		{0x43 /*0b1000011*/, 7},
		{0x42 /*0b1000010*/, 7},
		{0x83 /*0b10000011*/, 8},
		{0x82 /*0b10000010*/, 8},
		{0x81 /*0b10000001*/, 8},
		{0x80 /*0b10000000*/, 8}
};


/**
 * Compress
 * compresses the data in the buffer and writes the result to the same buffer.
 *
 * @param buffer	pointer to the buffer
 * @param insize	size of the input data
 *
 * @return	size of the compressed data
 */
size_t Compress(uint8_t* buffer, size_t insize) {
	size_t	inpos	= 0;
	size_t	outpos	= 0;

	uint8_t		*actualPointer, *historyPointer;

	uint8_t		maxoffset, offset, bestoffset;
	uint8_t		offsets[4];

	// The following variables are not only used for the LZMA coding,
	// but also for the Huffmancoding.
	uint16_t	maxlength, length, bestlength;

	static HUFFLIST	list[HUFF_LIST_LENGTH];
	code_t		code_sym	= 0;
	int8_t		code_length	= 0;

	static uint8_t		internBuffer[INTERN_BUFFER_LENGTH];
#if INTERN_BUFFER_LENGTH < 255
	uint8_t		internBufferWrite	= 0;
	uint8_t		internBufferRead	= 0;
#else
	uint16_t		internBufferWrite	= 0;
	uint16_t		internBufferRead	= 0;
#endif

	uint8_t		i;	// used for several purposes

	// Initialize Symbollist
	for (i = 0; i < HUFF_LIST_LENGTH; i++) {
			list[i].count	= 0;
			//list[i].symbol	= 0;
	}

	// Initialize Offsets
	for (i = 0; i < 4; i++) {
			offsets[i] = 0;
	}

	while( inpos < insize ) {
    	maxoffset = ( inpos > LZ_MAX_OFFSET ) ?
    			LZ_MAX_OFFSET : inpos;
        maxlength = ( insize - inpos ) > LZ_MAX_LENGTH ?
        		LZ_MAX_LENGTH : (insize - inpos);

        actualPointer = &buffer[ inpos ];

        bestlength = 2;
        bestoffset = 0;

        // Search for the following string in the history

        // First search in the offset History (if defined):
#ifdef LZMH_FIRSTSEARCHHISTORY
        for( i = 0; i < 4; i++) {
        	offset = offsets[ i ];
        	if (offset != 0) {
				historyPointer = &actualPointer[ -(int)offset ];
				if( (actualPointer[ 0 ] == historyPointer[ 0 ]) &&
					(actualPointer[ bestlength ] == historyPointer[ bestlength ]) ) {

					for( length = 1;
								(length < maxlength) &&
										(actualPointer[length] == historyPointer[length]);
								++ length );

					if( length > bestlength ) {
						bestlength = length;
						bestoffset = offset;
					}
				}
        	}
        }
#endif
        // Then search the hole History
		for( offset = 1;
				(offset <= maxoffset) &&
						(bestlength < maxlength);
				++ offset ) {

			historyPointer = &actualPointer[ -(int)offset ];

			if( (actualPointer[ 0 ] == historyPointer[ 0 ]) &&
				(actualPointer[ bestlength ] == historyPointer[ bestlength ]) ) {

				for( length = 1;
							(length < maxlength) &&
									(actualPointer[length] == historyPointer[length]);
							++ length );

				if( length > bestlength ) {
					bestlength = length;
					bestoffset = offset;
				}
			}
		}

        if( bestlength >= 3 ) { // LZMA Coding

			// Write Offset
			if ( offsets[ 0 ] == bestoffset ) {
				code_length	+= 4;
				code_sym	|= 0b0110 << ( CODE_BUFFER - code_length );
			} else if ( offsets[ 1 ] == bestoffset ) {
				offsets[1] = offsets[0];
				offsets[0] = bestoffset;

				code_length	+= 5;
				code_sym	|= 0b01110 << ( CODE_BUFFER - code_length );
			} else if ( offsets[ 2 ] == bestoffset ) {
				offsets[2] = offsets[1];
				offsets[1] = offsets[0];
				offsets[0] = bestoffset;

				code_length += 6;
				code_sym	|= 0b011110 << ( CODE_BUFFER - code_length );
			} else if ( offsets[ 3 ] == bestoffset ) {
				offsets[3] = offsets[2];
				offsets[2] = offsets[1];
				offsets[1] = offsets[0];
				offsets[0] = bestoffset;

				code_length += 6;
				code_sym	|= 0b011111 << ( CODE_BUFFER - code_length );
			} else {
				offsets[3] = offsets[2];
				offsets[2] = offsets[1];
				offsets[1] = offsets[0];
				offsets[0] = bestoffset;

				code_length += 10;
				code_sym	|= ( 0b0100000000 | ( bestoffset - 1 ) ) <<
							( CODE_BUFFER - code_length );
			}

			// Write Length
			if ( bestlength < 11 ) { // 3 - 10
				code_length	+= 4;
				code_sym	|= ( bestlength - 3 ) <<
						( CODE_BUFFER - code_length );
			} else if (bestlength < 19) { // 11 - 18
				code_length += 5;
				code_sym	|= ( 0x10 | ( bestlength - 11 ) ) <<
						( CODE_BUFFER - code_length );
			} else { // 19 - LZ_MAX_LENGTH
				code_length += 10;
				code_sym	|= ( 0x300 | ( bestlength - 19 ) ) <<
						( CODE_BUFFER - code_length );
			}

			inpos += bestlength;

        } else { // ATH Coding
			length = 0;
			maxlength = HUFF_SYMBOL_NOT_FOUND;

			i = buffer[ inpos++ ];

        	// Search for the Symbol in the list
			while (length < HUFF_LIST_LENGTH && list[length].count > 0)  {
				if (list[length].symbol == i) {
					maxlength = length;
					if (list[length].count < ((1<<16)-1)) {
						bestlength = list[length].count + 1;

						// Bubble Sort
						while ((length > 0)	&& (bestlength > list[length-1].count)) {
							list[length].symbol = list[length-1].symbol;
							length--;
						}

						list[length].count = bestlength;
						list[length].symbol = i;
					}
					break;
				}
				length++;
			}

        	if (maxlength == HUFF_SYMBOL_NOT_FOUND && length < HUFF_LIST_LENGTH) {
				// Add Symbol to list
				list[length].symbol = i;
				list[length].count = 1;
        	}

        	if (maxlength < TREE_LENGHT) { // Huffman Code
        		code_length	+= tree[maxlength].length;
        		code_sym	|= tree[maxlength].code << (
        				CODE_BUFFER - code_length);
        	} else { // Bytecode
				code_length	+= HUFF_BYTE_LENGTH + 2;
				code_sym	|= i << (CODE_BUFFER -
						code_length);
        	}

        }

      	while ( code_length >= 8 ) {
      		internBuffer[internBufferWrite++] = (code_sym >> (CODE_BUFFER - 8));

      		if (internBufferWrite == INTERN_BUFFER_LENGTH) internBufferWrite = 0;

      		code_sym = code_sym << 8;
      		code_length -= 8;
      	}

      	if ( inpos > LZ_MAX_OFFSET) {
      		while ( internBufferRead != internBufferWrite && outpos < (inpos - LZ_MAX_OFFSET)) {
      			buffer[ outpos++ ] = internBuffer[internBufferRead++];

      			if (internBufferRead == INTERN_BUFFER_LENGTH) internBufferRead = 0;
      		}
      	}

    }

    // Write Last Byte if any
	while ( internBufferRead != internBufferWrite) {
		buffer[ outpos++ ] = internBuffer[internBufferRead++];
		if (internBufferRead == INTERN_BUFFER_LENGTH) internBufferRead = 0;
	}

    if (code_length > 0) {
		buffer[ outpos++ ] = (code_sym >> (CODE_BUFFER - 8)) ;
	}

    return outpos;
}

#ifndef NODECOMPRESSOR

/**
 * DeCompress
 * decompresses the data in the inbuffer and writes it to the outbuffer
 *
 * @param inbuffer	Pointer to the compressed data
 * @param insize	Size of the compressed data
 * @param outbuffer	Pointer to the output data
 * @param outsize	Maximum output size, default 0xFFFF
 *
 * @return	Size of the decompressed data or 0 on error
 */
size_t DeCompress(uint8_t* inbuffer, size_t insize, uint8_t* outbuffer, size_t outsize)
{
    size_t		i, inpos, outpos, length, offset;

    uint16_t	offsets[4];

	HUFFLIST	list[HUFF_LIST_LENGTH];
	code_t		code_sym	= 0;
	int8_t		code_length	= 0;

    // Do we have anything to uncompress?
    if( insize < 1 ) {
        return( 0 );
    }

	for (i = 0; i < HUFF_LIST_LENGTH; i++) {
			list[i].count		= 0;
			//list[i].symbol	= 0;
	}
	for (i = 0; i < 4; i++) {
    		offsets[i] = 0;
    }

    inpos = 0;

    // Main decompression loop
    outpos = 0;
    do {
		while (inpos < insize && code_length < 24) {
				code_length += 8;
				code_sym |= inbuffer[ inpos ++ ] << (32 - code_length);
		}
		if ((code_sym & 0x80000000) != 0) {
			// Huffman Code
			for (i = 0; i < TREE_LENGHT; i++) {
				if ((code_length >= tree[i].length) &&
						(code_sym >> (32 - tree[i].length)) == tree[i].code) {

					code_length -= tree[i].length;
					code_sym = code_sym << tree[i].length;

					outbuffer[ outpos++ ] = list[i].symbol;

					if (list[i].count < ((1<<16)-1)) {
						length = list[i].count + 1;

						// Bubble Sort
						while ((i > 0)	&& (length > list[i-1].count)) {
							list[i].symbol = list[i-1].symbol;
							i--;
						}

						list[i].count = length;
						list[i].symbol = outbuffer[ (outpos-1) ];
					}

					break;
				}
			}
			if (i == TREE_LENGHT) {
				return 0;
			}
		} else if ((code_sym & 0x40000000) == 0) {
			// Prefix Code
			i = (code_sym >> (32 - (HUFF_BYTE_LENGTH + 2))) & ((1 << HUFF_BYTE_LENGTH)-1);
			code_length -= (HUFF_BYTE_LENGTH + 2);
			code_sym = code_sym << (HUFF_BYTE_LENGTH + 2);

			outbuffer[ outpos++ ] = i;
			for (length = 0; length < HUFF_LIST_LENGTH && list[length].count > 0 && list[length].symbol != i; length++);
			if (length < HUFF_LIST_LENGTH && list[length].count < ((1<<16)-1)) {
				// Because we don't need an offset here, we use the variable
				// for saving the count.
				offset = list[length].count + 1;

				// Bubble Sort
				while ((length > 0)	&& (offset > list[length-1].count)) {
					list[length] = list[length-1];
					length--;
				}

				list[length].count = offset;
				list[length].symbol = i;
			}
		} else {
			// LZMA Code

			code_length -= 2;
			code_sym = code_sym << 2;

			// Offset
			if ((code_sym & 0x80000000) == 0) {
					offset = (code_sym >> 24) + 1;
					code_length -= 8;
					code_sym = code_sym << 8;
					offsets[3] = offsets[2];
					offsets[2] = offsets[1];
					offsets[1] = offsets[0];
					offsets[0] = offset;
			} else {
					code_length -= 1;
					code_sym = code_sym << 1;
					if ((code_sym & 0x80000000) == 0) {
							offset = offsets[0];
					} else {
							code_length -= 1;
							code_sym = code_sym << 1;
							if ((code_sym & 0x80000000) == 0) {
									offset = offsets[1];
									offsets[1] = offsets[0];
									offsets[0] = offset;
							} else {
									code_length -= 1;
									code_sym = code_sym << 1;
									if ((code_sym & 0x80000000) == 0) {
											offset = offsets[2];
											offsets[2] = offsets[1];
											offsets[1] = offsets[0];
											offsets[0] = offset;
									} else {
											offset = offsets[3];
											offsets[3] = offsets[2];
											offsets[2] = offsets[1];
											offsets[1] = offsets[0];
											offsets[0] = offset;
									}
							}
					}
					code_length -= 1;
					code_sym = code_sym << 1;
			}

			// Length
			if ((code_sym & 0x80000000) == 0) {
					length = ((code_sym >> 28) & 0x07) + 3;
					code_length -= 4;
					code_sym = code_sym << 4;
			} else {
					code_length -= 1;
					code_sym = code_sym << 1;
					if ((code_sym & 0x80000000) == 0) {
							length = ((code_sym >> 28) & 0x7) + 11;
							code_length -= 4;
							code_sym = code_sym << 4;
					} else {
							length = ((code_sym >> 23) & 0xFF) + 19;
							code_length -= 9;
							code_sym = code_sym << 9;
					}
			}
			for( i = 0; i < length; ++ i ) {
				outbuffer[ outpos ] = outbuffer[ outpos - offset ];
				++ outpos;
			}
		}
	} while( (outpos < outsize) && (inpos < insize || code_sym > 0));

    return outpos;
}

#endif
